
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftORleftANDAND CONTAINS EQ HAS IDENTIFIER_BOOL IDENTIFIER_IP IDENTIFIER_LIST IDENTIFIER_NUM IDENTIFIER_STR IN LPAREN NEQ NOT_CONTAINS NOT_HAS NOT_IN OR RPAREN VALUE_BOOL VALUE_IP VALUE_LIST_IP_CIDR VALUE_LIST_NUM VALUE_LIST_STR VALUE_NUM VALUE_STRexpression : IDENTIFIER_STR EQ VALUE_STR\n    | IDENTIFIER_STR NEQ VALUE_STR\n    | IDENTIFIER_STR CONTAINS VALUE_STR\n    | IDENTIFIER_STR NOT_CONTAINS VALUE_STR\n    | IDENTIFIER_STR IN VALUE_LIST_STR\n    | IDENTIFIER_STR NOT_IN VALUE_LIST_STR\n    | IDENTIFIER_LIST HAS VALUE_STR\n    | IDENTIFIER_LIST NOT_HAS VALUE_STR\n    | IDENTIFIER_IP EQ VALUE_IP\n    | IDENTIFIER_IP NEQ VALUE_IP\n    | IDENTIFIER_IP IN VALUE_LIST_IP_CIDR\n    | IDENTIFIER_IP NOT_IN VALUE_LIST_IP_CIDR\n    | IDENTIFIER_NUM EQ VALUE_NUM\n    | IDENTIFIER_NUM NEQ VALUE_NUM\n    | IDENTIFIER_NUM IN VALUE_LIST_NUM\n    | IDENTIFIER_NUM NOT_IN VALUE_LIST_NUM\n    | IDENTIFIER_BOOL EQ VALUE_BOOLexpression : expression AND expression\n    | expression OR expressionexpression : LPAREN expression RPAREN'
    
_lr_action_items = {'IDENTIFIER_STR':([0,7,8,9,],[2,2,2,2,]),'IDENTIFIER_LIST':([0,7,8,9,],[3,3,3,3,]),'IDENTIFIER_IP':([0,7,8,9,],[4,4,4,4,]),'IDENTIFIER_NUM':([0,7,8,9,],[5,5,5,5,]),'IDENTIFIER_BOOL':([0,7,8,9,],[6,6,6,6,]),'LPAREN':([0,7,8,9,],[7,7,7,7,]),'$end':([1,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,],[0,-18,-19,-1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12,-13,-14,-15,-16,-17,-20,]),'AND':([1,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,],[8,8,-18,8,-1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12,-13,-14,-15,-16,-17,-20,]),'OR':([1,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,],[9,9,-18,-19,-1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12,-13,-14,-15,-16,-17,-20,]),'EQ':([2,4,5,6,],[10,18,22,26,]),'NEQ':([2,4,5,],[11,19,23,]),'CONTAINS':([2,],[12,]),'NOT_CONTAINS':([2,],[13,]),'IN':([2,4,5,],[14,20,24,]),'NOT_IN':([2,4,5,],[15,21,25,]),'HAS':([3,],[16,]),'NOT_HAS':([3,],[17,]),'VALUE_STR':([10,11,12,13,16,17,],[30,31,32,33,36,37,]),'VALUE_LIST_STR':([14,15,],[34,35,]),'VALUE_IP':([18,19,],[38,39,]),'VALUE_LIST_IP_CIDR':([20,21,],[40,41,]),'VALUE_NUM':([22,23,],[42,43,]),'VALUE_LIST_NUM':([24,25,],[44,45,]),'VALUE_BOOL':([26,],[46,]),'RPAREN':([27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,],[47,-18,-19,-1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12,-13,-14,-15,-16,-17,-20,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,7,8,9,],[1,27,28,29,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> IDENTIFIER_STR EQ VALUE_STR','expression',3,'p_expression_unit','parser.py',126),
  ('expression -> IDENTIFIER_STR NEQ VALUE_STR','expression',3,'p_expression_unit','parser.py',127),
  ('expression -> IDENTIFIER_STR CONTAINS VALUE_STR','expression',3,'p_expression_unit','parser.py',128),
  ('expression -> IDENTIFIER_STR NOT_CONTAINS VALUE_STR','expression',3,'p_expression_unit','parser.py',129),
  ('expression -> IDENTIFIER_STR IN VALUE_LIST_STR','expression',3,'p_expression_unit','parser.py',130),
  ('expression -> IDENTIFIER_STR NOT_IN VALUE_LIST_STR','expression',3,'p_expression_unit','parser.py',131),
  ('expression -> IDENTIFIER_LIST HAS VALUE_STR','expression',3,'p_expression_unit','parser.py',132),
  ('expression -> IDENTIFIER_LIST NOT_HAS VALUE_STR','expression',3,'p_expression_unit','parser.py',133),
  ('expression -> IDENTIFIER_IP EQ VALUE_IP','expression',3,'p_expression_unit','parser.py',134),
  ('expression -> IDENTIFIER_IP NEQ VALUE_IP','expression',3,'p_expression_unit','parser.py',135),
  ('expression -> IDENTIFIER_IP IN VALUE_LIST_IP_CIDR','expression',3,'p_expression_unit','parser.py',136),
  ('expression -> IDENTIFIER_IP NOT_IN VALUE_LIST_IP_CIDR','expression',3,'p_expression_unit','parser.py',137),
  ('expression -> IDENTIFIER_NUM EQ VALUE_NUM','expression',3,'p_expression_unit','parser.py',138),
  ('expression -> IDENTIFIER_NUM NEQ VALUE_NUM','expression',3,'p_expression_unit','parser.py',139),
  ('expression -> IDENTIFIER_NUM IN VALUE_LIST_NUM','expression',3,'p_expression_unit','parser.py',140),
  ('expression -> IDENTIFIER_NUM NOT_IN VALUE_LIST_NUM','expression',3,'p_expression_unit','parser.py',141),
  ('expression -> IDENTIFIER_BOOL EQ VALUE_BOOL','expression',3,'p_expression_unit','parser.py',142),
  ('expression -> expression AND expression','expression',3,'p_expression_binop','parser.py',159),
  ('expression -> expression OR expression','expression',3,'p_expression_binop','parser.py',160),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','parser.py',167),
]



import re
from sharh.expr import Literal, Conjunction, Disjunction


class ParseError(Exception):
    pass


class ExpressionTree:
    def __init__(self):
        self.tree = []
        self.stack = []
        self.expressions = {}

        self.original_expr_was_dnf = True

    def push(self, expr_args):
        self.stack.append(Literal(*expr_args))

    def commit(self, operator):
        right = self.stack.pop()
        left = self.stack.pop()

        if operator == t_AND:
            if (
                (isinstance(left, Literal) or isinstance(left, Conjunction))
                and isinstance(right, Disjunction)
            ) or (
                (isinstance(right, Literal) or isinstance(right, Conjunction))
                and isinstance(left, Disjunction)
            ):
                self.original_expr_was_dnf = False

            self.stack.append(left * right)
        elif operator == t_OR:
            self.stack.append(left + right)


tree = ExpressionTree()


tokens = (
    "IDENTIFIER_STR",
    "VALUE_STR",
    "VALUE_LIST_STR",
    "IDENTIFIER_LIST",
    "IDENTIFIER_IP",
    "VALUE_IP",
    "VALUE_LIST_IP_CIDR",
    "IDENTIFIER_NUM",
    "VALUE_NUM",
    "VALUE_LIST_NUM",
    "IDENTIFIER_BOOL",
    "VALUE_BOOL",
    "EQ",
    "NEQ",
    "IN",
    "NOT_IN",
    "HAS",
    "NOT_HAS",
    "CONTAINS",
    "NOT_CONTAINS",
    "AND",
    "OR",
    "LPAREN",
    "RPAREN",
)

IP_OR_CIDR_PATTERN = (
    r"((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)\.?\b){4}(\/([1-9]|[1-2]\d|3[0-2]))?"
)

t_IDENTIFIER_STR = (
    r"ip.geoip.country|ip.geoip.continent|"
    r"http.method|http.version|http.headers.user_agent|http.headers.x_forwarded_for|http.headers.referer|"
    r"http.headers\['[a-zA-Z0-9\-\_]+'\]|"
    r"device"
)
t_VALUE_STR = r"'[a-zA-Z0-9\/\-\.:\s]+'"
t_VALUE_LIST_STR = rf"\[\s*({t_VALUE_STR}\s*,\s*)*{t_VALUE_STR}\s*\]"
t_IDENTIFIER_LIST = r"http.headers"
t_IDENTIFIER_IP = "ip.addr"
t_VALUE_IP = r"((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)\.?\b){4}"
t_VALUE_LIST_IP_CIDR = rf"\[\s*({IP_OR_CIDR_PATTERN}\s*,\s*)*{IP_OR_CIDR_PATTERN}\s*\]"
t_IDENTIFIER_NUM = "ip.geoip.asn"
t_VALUE_NUM = "[0-9]{1,11}"
t_VALUE_LIST_NUM = rf"\[\s*({t_VALUE_NUM}\s*,\s*)*{t_VALUE_NUM}\s*\]"
t_IDENTIFIER_BOOL = r"http.secure"
t_VALUE_BOOL = r"(true)|(false)"
t_EQ = r"=="
t_NEQ = r"!="
t_IN = r"in"
t_NOT_IN = r"not\ in"
t_HAS = r"has"
t_NOT_HAS = r"not\ has"
t_CONTAINS = r"contains"
t_NOT_CONTAINS = r"not\ contains"
t_AND = r"and"
t_OR = r"or"
t_LPAREN = r"\("
t_RPAREN = r"\)"

# Ignored characters
t_ignore = " \t\n"

precedence = (
    ("left", "OR"),
    ("left", "AND"),
)


def t_error(t):
    t.lexer.skip(1)


import ply.lex as lex

lexer = lex.lex()


def p_error(t):
    raise ParseError()


def p_expression_unit(t):
    """expression : IDENTIFIER_STR EQ VALUE_STR
    | IDENTIFIER_STR NEQ VALUE_STR
    | IDENTIFIER_STR CONTAINS VALUE_STR
    | IDENTIFIER_STR NOT_CONTAINS VALUE_STR
    | IDENTIFIER_STR IN VALUE_LIST_STR
    | IDENTIFIER_STR NOT_IN VALUE_LIST_STR
    | IDENTIFIER_LIST HAS VALUE_STR
    | IDENTIFIER_LIST NOT_HAS VALUE_STR
    | IDENTIFIER_IP EQ VALUE_IP
    | IDENTIFIER_IP NEQ VALUE_IP
    | IDENTIFIER_IP IN VALUE_LIST_IP_CIDR
    | IDENTIFIER_IP NOT_IN VALUE_LIST_IP_CIDR
    | IDENTIFIER_NUM EQ VALUE_NUM
    | IDENTIFIER_NUM NEQ VALUE_NUM
    | IDENTIFIER_NUM IN VALUE_LIST_NUM
    | IDENTIFIER_NUM NOT_IN VALUE_LIST_NUM
    | IDENTIFIER_BOOL EQ VALUE_BOOL"""

    identifier = t[1]
    operation = t[2]
    value = t[3]

    if re.match(t_VALUE_STR, value):
        value = value[1:-1]
    elif re.match(t_VALUE_BOOL, value):
        value = True if value == "true" else False

    try:
        tree.push([identifier, operation, value])
    except ValueError as e:
        raise ParseError(*e.args)
    t[0] = [[t[1], t[2], t[3]]]


def p_expression_binop(t):
    """expression : expression AND expression
    | expression OR expression"""

    if t[2] in [t_AND, t_OR]:
        tree.commit(t[2])


def p_expression_group(t):
    "expression : LPAREN expression RPAREN"
    t[0] = t[2]


import ply.yacc as yacc

parser = yacc.yacc()


def parse(s):
    if not s.strip():
        return Disjunction([])

    global tree
    parser.parse(s)

    try:
        parsed = tree.stack.pop()
    except IndexError:
        raise ParseError()

    if isinstance(parsed, Literal):
        parsed = Conjunction([parsed])
    if isinstance(parsed, Conjunction):
        parsed = Disjunction([parsed])

    parsed.original_expr_was_dnf = tree.original_expr_was_dnf
    tree = ExpressionTree()

    return parsed



import pytest
from sharh.parser import parse, ParseError

def test_string_identifier_eq():
    tree = parse("http.headers.user_agent == 'curl/8.1-beta'")

    assert tree.to_expr_notation() == ["http_user_agent", "==", "curl/8.1-beta"]

    # with space
    tree = parse("http.headers.user_agent == 'curl/8.1-beta test'")
    assert tree.to_expr_notation() == ["http_user_agent", "==", "curl/8.1-beta test"]

def test_string_identifier_neq():
    tree = parse("http.headers.user_agent != 'curl/8.1-beta'")

    assert tree.to_expr_notation() == ["http_user_agent", "~=", "curl/8.1-beta"]

def test_string_identifer_contains():
    tree = parse("http.headers.user_agent contains 'curl/8.1-beta'")

    assert tree.to_expr_notation() == ["http_user_agent", "~~", "curl/8.1-beta"]

def test_string_identifier_not_contains():
    tree = parse("http.headers.user_agent not contains 'curl/8.1-beta'")

    assert tree.to_expr_notation() == ["http_user_agent", "!", "~~", "curl/8.1-beta"]

def test_string_identifier_in():
    tree = parse("http.headers.user_agent in [ 'curl', 'firefox'  ,'chrome'  ]")
    assert tree.to_expr_notation() == ["http_user_agent", "in", ['curl', 'firefox', 'chrome']]

def test_string_identifier_not_in():
    tree = parse("http.headers.user_agent not in ['curl' ,'firefox',  'chrome' ]")
    assert tree.to_expr_notation() == ["http_user_agent", "!", "in", ['curl', 'firefox', 'chrome']]

def test_string_identifier_dont_support_has():
    with pytest.raises(ParseError):
        parse("http.headers.user_agent has 'curl/8.1-beta'")

def test_custom_header():
    tree = parse("http.headers['content-type'] == 'text/html'")

    assert tree.to_expr_notation() == ["http_content_type", "==", "text/html"]

def test_list_identifier_has():
    tree = parse("http.headers has 'test-key'")

    assert tree.to_expr_notation() == ["headers", "has", "test-key"]

def test_list_identifier_has_not():
    tree = parse("http.headers not has 'test-key'")

    assert tree.to_expr_notation() == ["headers", "!", "has", "test-key"]

def test_asn():
    tree = parse("ip.geoip.asn == 1234")

    assert tree.to_expr_notation() == ["geoip2_data_asn", "==", "1234"]

def test_asn_invalid():
    with pytest.raises(ParseError):
        parse("ip.geoip.asn == '1234'")

def test_asn_in():
    tree = parse("ip.geoip.asn in [  1234 , 345 ]")
    assert tree.to_expr_notation() == ["geoip2_data_asn", "in", ["1234", "345"]]

def test_ip():
    tree = parse("ip.addr == 127.0.0.1")

    assert tree.to_expr_notation() == ["remote_addr", "==", "127.0.0.1"]

def test_ip_invalid():
    with pytest.raises(ParseError):
        parse("ip.addr == 256.1.1.1")

def test_ip_in():
    tree = parse("ip.addr in [ 192.168.1.1/28   ,  127.0.0.1 ]")

    assert tree.to_expr_notation() == ["remote_addr", "ipmatch", ["192.168.1.1/28", "127.0.0.1"]]

def test_ip_not_in():
    tree = parse("ip.addr not in [192.168.1.1/28, 127.0.0.1]")

    assert tree.to_expr_notation() == ["remote_addr", "!", "ipmatch", ["192.168.1.1/28", "127.0.0.1"]]

def test_boolean_variables():
    tree = parse("http.secure == true")
    assert tree.to_expr_notation() == ["scheme", "==", "https"]

    tree = parse("http.secure == false")
    assert tree.to_expr_notation() == ["scheme", "==", "http"]

    with pytest.raises(ParseError):
        parse("http.secure == 'invalid value")


def test_multiple():
    tree = parse("http.headers.user_agent != 'curl' and ip.addr == 127.0.0.1")

    assert tree.to_expr_notation() == ["AND", ["http_user_agent", "~=", "curl"], ["remote_addr", "==", "127.0.0.1"]]


def test_additional_var_mapping():
    tree = parse("http.headers.referer == 'https://github.com'")

    assert tree.to_expr_notation() == ["http_referer", "==", "https://github.com"]

    tree.set_additional_var_mapping({"http.headers.referer": "http_referrer"})

    assert tree.to_expr_notation() == ["http_referrer", "==", "https://github.com"]



from __future__ import annotations

from functools import reduce
from typing import Dict, List

VAR_MAPPING = {
    "http.method": "request_method",
    "http.version": "server_protocol",
    "ip.geoip.country": "geoip2_data_country_iso_code",
    "ip.geoip.continent": "geoip2_data_continent_code",
    "ip.geoip.asn": "geoip2_data_asn",
    "ip.addr": "remote_addr",
    "http.headers": "headers",
    "http.headers.user_agent": "http_user_agent",
    "http.headers.x_forwarded_for": "http_x_forwarded_for",
    "http.headers.referer": "http_referer",
    "http.secure": "scheme",
}


def add_literal_literal(l1: Literal, l2: Literal):
    return Disjunction([Conjunction([l1]), Conjunction([l2])])


def add_literal_conjunction(
    literal: Literal, conjunction: Conjunction, literal_first=True
):
    if literal_first:
        return Disjunction([Conjunction([literal]), conjunction])
    return Disjunction([conjunction, Conjunction([literal])])


def add_literal_disjunction(
    literal: Literal, disjunction: Disjunction, literal_first=True
):
    if literal_first:
        return Disjunction([Conjunction([literal]), *disjunction.conjunctions])
    return Disjunction([*disjunction.conjunctions, Conjunction([literal])])


def add_conjunction_conjunction(c1: Conjunction, c2: Conjunction):
    return Disjunction([c1, c2])


def add_conjunction_disjunction(
    conjunction: Conjunction, disjunction: Disjunction, conjunction_first=True
):
    if conjunction_first:
        return Disjunction([conjunction, *disjunction.conjunctions])
    return Disjunction([*disjunction.conjunctions, conjunction])


def add_disjunction_disjunction(d1: Disjunction, d2: Disjunction):
    return Disjunction([*d1.conjunctions, *d2.conjunctions])


def mul_literal_literal(l1: Literal, l2: Literal):
    return Conjunction([l1, l2])


def mul_literal_conjunction(
    literal: Literal, conjunction: Conjunction, literal_first=True
):
    if literal_first:
        return Conjunction([literal, *conjunction.literals])
    return Conjunction([*conjunction.literals, literal])


def mul_literal_disjunction(
    literal: Literal, disjunction: Disjunction, literal_first=True
):
    return Disjunction(
        list(
            map(
                lambda c: mul_literal_conjunction(
                    literal=literal, conjunction=c, literal_first=literal_first
                ),
                disjunction.conjunctions,
            )
        )
    )


def mul_conjunction_conjunction(c1: Conjunction, c2: Conjunction):
    return Conjunction([*c1.literals, *c2.literals])


def mul_conjunction_disjunction(
    conjunction: Conjunction, disjunction: Disjunction, conjunction_first=True
):
    if conjunction_first:
        return Disjunction(
            list(
                map(
                    lambda d: mul_conjunction_conjunction(conjunction, d),
                    disjunction.conjunctions,
                )
            )
        )

    return Disjunction(
        list(
            map(
                lambda d: mul_conjunction_conjunction(d, conjunction),
                disjunction.conjunctions,
            )
        )
    )


def mul_disjunction_disjunction(d1: Disjunction, d2: Disjunction):
    return reduce(
        lambda i, j: Disjunction([*i.conjunctions, *j.conjunctions]),
        map(
            lambda c: mul_conjunction_disjunction(c, d1, conjunction_first=False),
            d2.conjunctions,
        ),
        Disjunction([]),
    )


class DSLOperators:
    EQ = "=="
    NEQ = "!="
    HAS = "has"
    NOT_HAS = "not has"
    CONTAINS = "contains"
    NOT_CONTAINS = "not contains"
    IN = "in"
    NOT_IN = "not in"


class EXPROperators:
    NOT = "!"
    EQ = "=="
    NEQ = "~="
    HAS = "has"
    CONTAINS = "~~"
    IN = "in"
    IP_IN = "ipmatch"


class Literal:
    EXPR_OP = {
        DSLOperators.EQ: [EXPROperators.EQ],
        DSLOperators.NEQ: [EXPROperators.NEQ],
        DSLOperators.HAS: [EXPROperators.HAS],
        DSLOperators.NOT_HAS: [EXPROperators.NOT, EXPROperators.HAS],
        DSLOperators.CONTAINS: [EXPROperators.CONTAINS],
        DSLOperators.NOT_CONTAINS: [EXPROperators.NOT, EXPROperators.CONTAINS],
        DSLOperators.IN: [EXPROperators.IN],
        DSLOperators.NOT_IN: [EXPROperators.NOT, EXPROperators.IN],
    }

    def __init__(self, left, op, right):
        self.op = op
        self.left = left
        self.right = right

        self.additional_var_mapping = {}

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return str([self.left, self.op, self.right])

    def __repr__(self):
        return f"Literal <{str(self)}>"

    def __eq__(self, other):
        return str(self) == str(other)

    def __mul__(self, other):
        if isinstance(other, Literal):
            return mul_literal_literal(self, other)
        if isinstance(other, Conjunction):
            return mul_literal_conjunction(self, other)
        if isinstance(other, Disjunction):
            return mul_literal_disjunction(self, other)

    def __add__(self, other):
        if isinstance(other, Literal):
            return add_literal_literal(self, other)
        if isinstance(other, Conjunction):
            return add_literal_conjunction(self, other)
        if isinstance(other, Disjunction):
            return add_literal_disjunction(self, other)

    def set_additional_var_mapping(self, additional_var_mapping: Dict[str, str]):
        self.additional_var_mapping = additional_var_mapping

    def get_op(self):
        if self.left == "ip.addr":
            if self.op == DSLOperators.IN:
                return [EXPROperators.IP_IN]
            elif self.op == DSLOperators.NOT_IN:
                return [EXPROperators.NOT, EXPROperators.IP_IN]

        return self.EXPR_OP[self.op]

    def get_lvalue(self):
        v = self.additional_var_mapping.get(self.left)
        if v:
            return v

        v = VAR_MAPPING.get(self.left)
        if v:
            return v

        elif self.left.startswith("http.headers["):
            header_key = (
                self.left[len("http.headers") :].strip("[]'.").replace("-", "_")
            )
            return "http_" + header_key

        return self.left

    def get_rvalue(self):
        if self.op in (DSLOperators.IN, DSLOperators.NOT_IN):
            return list(map(lambda i: i.strip("' "), self.right.strip("[]").split(",")))

        elif self.left == "http.secure":
            return "https" if self.right == True else "http"

        return self.right

    def to_expr_notation(self):
        return [self.get_lvalue(), *self.get_op(), self.get_rvalue()]


class Conjunction:
    def __init__(self, literals: List[Literal]):
        self.literals = literals

    def __str__(self):
        return f"( {' and '.join(map(lambda i: str(i), self.literals))} )"

    def __repr__(self):
        return f"Conjunction <{str(self)}>"

    def __mul__(self, other):
        if isinstance(other, Literal):
            return mul_literal_conjunction(other, self, literal_first=False)
        if isinstance(other, Conjunction):
            return mul_conjunction_conjunction(self, other)
        if isinstance(other, Disjunction):
            return mul_conjunction_disjunction(self, other)

    def __add__(self, other):
        if isinstance(other, Literal):
            return add_literal_conjunction(other, self, literal_first=False)
        if isinstance(other, Conjunction):
            return add_conjunction_conjunction(self, other)
        if isinstance(other, Disjunction):
            return add_conjunction_disjunction(self, other)

    def set_additional_var_mapping(self, additional_var_mapping: Dict[str, str]):
        for l in self.literals:
            l.set_additional_var_mapping(additional_var_mapping)

    def to_expr_notation(self):
        if len(self.literals) == 0:
            return []
        elif len(self.literals) == 1:
            return self.literals[0].to_expr_notation()

        return ["AND", *map(lambda literal: literal.to_expr_notation(), self.literals)]


class Disjunction:
    def __init__(self, conjunctions: List[Conjunction]):
        self.conjunctions = conjunctions
        self.original_expr_was_dnf = True

    def __repr__(self):
        return f"Disjunction <{str(self)}>"

    def __str__(self):
        return f"( {' OR '.join(map(lambda i: str(i), self.conjunctions))} )"

    def __mul__(self, other):
        if isinstance(other, Literal):
            return mul_literal_disjunction(other, self, literal_first=False)
        if isinstance(other, Conjunction):
            return mul_conjunction_disjunction(other, self, conjunction_first=False)
        if isinstance(other, Disjunction):
            return mul_disjunction_disjunction(self, other)

    def __add__(self, other):
        if isinstance(other, Literal):
            return add_literal_disjunction(other, self, literal_first=False)
        if isinstance(other, Conjunction):
            return add_conjunction_disjunction(other, self, conjunction_first=False)
        if isinstance(other, Disjunction):
            return add_disjunction_disjunction(self, other)

    def set_additional_var_mapping(self, additional_var_mapping: Dict[str, str]):
        for conj in self.conjunctions:
            conj.set_additional_var_mapping(additional_var_mapping)

    def to_expr_notation(self):
        if len(self.conjunctions) == 0:
            return []
        elif len(self.conjunctions) == 1:
            return self.conjunctions[0].to_expr_notation()

        return [
            "OR",
            *map(lambda conjunction: conjunction.to_expr_notation(), self.conjunctions),
        ]
